// Element's fader channels data and functions

var ch_preview={grmode: false, grkey: -1};
const chnArmedStyle=20; // Wink style

/*
** chnInitTable
**
** initialize channel objects
*/
function chnInitTable()
{
	var n;

	ch_data=[];
	for(n=1; n<=64; n++)
	{
		ch_data[n]=
		{
			id:   n,			// channel number
			tag:  host.sprintf("mCH%02d", n),
			exists:  false, // true if channel is active
			pvmo: 1,			// preview mode: 1 - normal; 2 - ON -> Pvw OFF; 3 - Pvw ON -> OFF; 4 - both 2 and 3
			enbl:					// controls lock map: false - locked; true - unlocked
			{
				pgm: {1:true, 2:true, 3:true, 4:true},
				options:true,
				preview:true,
				onoff:  true,
				fader:  true,
				tbck:   true,
				hpsrc:  true
			},
			pgm: {1:0, 2:0, 3:0, 4:0}, // pgm1..pgm4
			srci: src_NONE_id, // source id
			src_name: "", // source name
			src_next: "", // next source name
			src_label:"", // source label
			lbl_dspl: 0,	// label display mode: 0 - never, 1 - always, 2 - alternate (3 seconds each)
			srct: 0,			// source type

		// Channel Rotary control
			knob_fn: knobFn.none,	// knob function
			knob_to: null,				// knob timeout
			gain: 0,					// fader trim gain (-25.0 .. 25.0 dB)
			src_gain: 0,			// source gain (in nodes) (-12.0 .. 12.0 dB - analog) (+18.0 .. +83.0 dB - mic)
			gain_lock: false,	// gain lock
			src_lwrp: false,	// is source added to LWRP control (source is added by lpid)

		// phone indications
			phone: false,     // Is this channel related to phone
			ph_icon: 0,       // Phone Icon to show (overrides default channel icons)
			ph_text: "",      // Phone text (overrides default channel text)
			ph_onair: false,  // Is phone on air now

			fader:      false,	// Fader state DOWN: false, UP: true
			groupStart: false,	// Group start: false: independent, 1: master, 2: slave
			faderStart: 0,			// Fader start Mode: 0 - OFF, 1 - normal, 2 - with Arm
			armed:      false,	// Channel is in armed state (OFF, but ready to be turned ON upon fader-up event)
			rden:       false,	// GPIO ready enabled
			rdcm:       false,	// GPIO ready command (valid only when rden = true)
			onsw:       false,	// if channel is ON or OFF
			mute:       false,	// if channel is muted
			preview:    false,	// if channel is in preview
			talkCR:     false,	// if talk to CR is active
			talkback:   false,	// talkback
			tbck_latch: false,	// talkback latch
			pview_key:  0,			// preview key: 0 - not pressed; 1 - pressed (pview was OFF); 2 - pressed (pview was ON)
			dplay_stat: 0,			// display stat
			allowAB:    false,	// allow AB switch

			recOnswLock: false, // ON/OFF lock while in rec_mode

			lpid: 0,	// if remote control of this source is enabled in source profile, then primary source id(PSID), otherwise 0
			hptg: "",	// current tag for HP selection
			hptg_scroll: "",		// current tag in scroll list (if we are in scroll mode)
			hp_scroll_mode: 0,	// 0=>volume, 1=>main source, 2=>preset's source
			hp_gain: 0,					// hp gain
			hp_src_name: "",
			gainconfirmTimeout: null,
			srcresetTimeout: null,
			hp_pset1: "PGM1",
			hp_pset2: "PGM2",
			presetTimeout: null,
			presetMode: 0, // 0 - none; 1 - configuring preset1; 2 - configuring preset2

			src_status: sstat.unknown,
		// Source sharing: owner
			src_owner_name:  "",
			src_owner_fader: 0
		};
	}

	ret_data=[];
	for(n=1; n<=2; n++)
	{
		ret_data[n]=
		{
			id:   n,			// channel number
			tag:  host.sprintf("mRET%d", n),

			pgm: {1:0, 2:0, 3:0, 4:0}, // pgm1..pgm4
			onsw:       false
		};
	}

// launch label alternation thread
	host.setTimeout(1000, chnAlternateLabels, null);
}

/*
** chnInitMask
*/
function chnInitMask(mask)
{
	var n, len;
	var cmsk="";

	len=mask.length;
	for(n=1; n<=64; n++)
	{
		if(n<=len && mask.charCodeAt(n-1))
		{
			ch_data[n].exists=true;
			mod.setChnKeyLed(n, keys.ChnPreview, 0);
			mod.setChnKeyLed(n, keys.ChnOptions, 0);
		}
		else
			ch_data[n].exists=false;
		cmsk+=ch_data[n].exists?"1":"0";
	}
	return cmsk;
}

/*
** chnUpdateFaderLevels
*/
function chnUpdateFaderLevels()
{
	var n;

	for(n=1; n<=64; n++)
		if(ch_data[n].exists)
			host.sendFaderLevel(n, mod.getChnFader(n));
}

/*
** setChnSrcName
*/
function setChnSrcName(chn, srcn, nxtn)
{
	srcn=host.strcclamp(srcn, 10);
	if(nxtn==null || nxtn=="")
		mod.setChnText(chn, srcn);
	else
	{
		nxtn=host.strcclamp(nxtn, 10);
		mod.setChnText(chn, "11--11-00000000-:"+srcn+nxtn, mod.BLINK);
	}
}

/*
** updateChnIcon
*/
function updateChnIcon(chn)
{
/* phone icon has a priority */
	if(ch_data[chn].ph_icon!=0)
	{
		mod.setChnIcon(chn, ch_data[chn].ph_icon);
		return;
	}
/* channel icon */
	switch(ch_data[chn].dplay_stat)
	{
	case 4: // HOST (Operator) talks
		mod.setChnIcon(chn, icons.HostTalk);
		break;
	case 5: // MUTE
		mod.setChnIcon(chn, icons.CrossBig);
		break;
	case 6: // talk to CR, GPIO remote talk
		mod.setChnIcon(chn, icons.TalkToCR);
		break;
	case 8: // talkback ON
		mod.setChnIcon(chn, icons.Talkback);
		break;
	case 128: // general error
	case 129: // disconnected output
	default:
  // Channel number or Listen only icon (when full control requested)
		mod.setChnIcon(chn, (ch_data[chn].src_status==sstat.locked_elsewhere)?icons.CrossSml:0);
		break;
	}
}

var displayLabelNow=false;	// alternates each 3 seconds

/*
** chnAlternateLabels
*/
function chnAlternateLabels()
{
	var n, chn;

	host.setTitle("Channel labels");
	while(1)
	{
		host.sleep(3000);
		displayLabelNow=true;
	// show labels on each channel where aplicable
		for(n=1; n<=64; n++)
		{
			chn=ch_data[n];
			if(!chn.exists) continue;

			if(chn.src_status in src_stat) continue;
			if(chn.lbl_dspl==2 && chn.src_next=="" && chn.ph_text=="" && chn.src_label!="")
				setChnSrcName(n, chn.src_label, "");
		}

		host.sleep(3000);
		displayLabelNow=false;
	// show source names on each channel where aplicable
		for(n=1; n<=64; n++)
		{
			chn=ch_data[n];
			if(!chn.exists) continue;

			if(chn.src_status in src_stat) continue;
			if(chn.lbl_dspl==2 && chn.src_next=="" && chn.ph_text=="" && chn.src_label!="")
				setChnSrcName(n, chn.src_name, "");
		}
	}
}

/*
** updateChnDisp
**
** updates channel display based on information from channel data
*/
function updateChnDisp(channel)
{
	var chn;

	if(!(channel in ch_data)) return;
	chn=ch_data[channel];

	if(chn.ph_text!="")
		setChnSrcName(channel, chn.ph_text, "");
	else if(chn.src_status in src_stat)
		mod.setChnText(channel, src_stat[chn.src_status], mod.SCROLL);
	else if(chn.lbl_dspl==0 || chn.src_next!="" || chn.src_label=="") // don't display label
		setChnSrcName(channel, chn.src_name, chn.src_next);
	else if(chn.lbl_dspl==1) // display label
		setChnSrcName(channel, chn.src_label, "");
	else if(chn.lbl_dspl==2) // alternation mode
		setChnSrcName(channel, displayLabelNow?chn.src_label:chn.src_name, "");
}

function setChnPhoneIcon(channel, icon)
{
	ch_data[channel].ph_icon=icon;
	updateChnIcon(channel);
}
function setChnPhoneText(channel, text)
{
	ch_data[channel].ph_text=text;
	updateChnDisp(channel);
}
function setChnPhoneOnAir(channel, onair)
{
	if(ch_data[channel].ph_onair==onair) return;
//	mod.setChnKeyLed(channel, keys.ChnSet, onair?255:0);
//	if(!onair) chnLineDroped(channel);
	ch_data[channel].ph_onair=onair;
}
function setChnPhoneLed(channel, style)
{
	mod.setChnKeyLed(channel, keys.ChnSet, style);
}

/*
** updateChannelOnOff
**
** updates channel ON/OFF lamps based on information from channel data
*/
function updateChannelOnOff(channel)
{
	var chn=ch_data[channel];

	mod.setChnKeyLed(channel, keys.ChnON , chn.onsw?(chn.mute?0:255):(chn.armed?chnArmedStyle:0));
	mod.setChnKeyLed(channel, keys.ChnOFF, (chn.onsw || (chn.rden && !chn.rdcm))?0:255);
}

/*
** chnIsPgmOn
*/
function chnIsPgmOn(pgm_nr)
{
	var n;

	for(n=1; n<=2; n++)
		if(ret_data[n].pgm[pgm_nr]==1)
			return true;
	for(n=1; n<=64; n++)
		if(ch_data[n].exists && ch_data[n].pgm[pgm_nr]==1)
			return true;
	return false;
}

/*
** chnIsPviewDown
*/
function chnIsPviewDown()
{
	var n;

	for(n=1; n<=64; n++)
		if(ch_data[n].exists && ch_data[n].pview_key!=0)
			return true;
	return false;
}

/*
** chnSetPview
*/
function chnSetPview(channel, on)
{
	host.sendChnCmsg(channel, {id:'WRIN', bPREV:on?1:0});
	if(!on && ch_data[channel].tbck_latch)
	{
		ch_data[channel].tbck_latch=false;
		if(!mod.getChnKey(channel, keys.ChnTalkback))
			host.sendChnCmsg(channel, {id:'WRIN', bTBCK:0});
	}
}

/*
** chnTurnOffAllPview
*/
function chnTurnOffAllPview()
{
	var n;

	for(n=1; n<=64; n++)
		if(ch_data[n].exists && ch_data[n].preview && !ch_data[n].tbck_latch)
			chnSetPview(n, false);
}

/*
** chnPreON
**
** checks if channel can be turned ON
*/
function chnPreON(channel)
{
	var chn=ch_data[channel];

	if(chn.recOnswLock && recmode_active) return false;
	if(chn.faderStart && !chn.fader)
	{
	// Don't turn channel ON if fader start and fader is down
		if(chn.faderStart==2 && !chn.onsw && !chn.armed)
		{
			chn.armed=true;
			updateChannelOnOff(channel);
		}
		return false;
	}
	return true;
}

/*
** chnPreOFF
*/
function chnPreOFF(channel)
{
	var chn=ch_data[channel];

	if(chn.recOnswLock && recmode_active) return false;
	if(chn.armed)
	{
	// disarm fader
		chn.armed=false;
		updateChannelOnOff(channel);
	}
	return true;
}

/*
** chnSetOnOff
*/
function chnSetOnOff(channel, on)
{
	var n;

// send command always - ACL will decide if it needs switching or no
// we need this to be able to send some GPIO commands when channel is already on and on cmd is sent
// send as a request, because ACL will send indication anyway, and we don't need two indications
	host.sendChnCmsgRq(channel, {id:"WRIN", bONSW:(on?1:0)});
// process group start
	if(ch_data[channel].groupStart==1)
	{
		for(n=1; n<=64; n++)
			if(n!=channel && ch_data[n].exists && ch_data[n].groupStart)
		{
			if(on)
			{
				if(chnPreON(n))
					host.sendChnCmsgRq(n, {id:"WRIN", bONSW:1});
			}
			else
			{
				if(chnPreOFF(n))
					host.sendChnCmsgRq(n, {id:"WRIN", bONSW:0});
			}
		}
	}
}

/*
** chnSetMute
*/
function chnSetMute(channel, on)
{
	host.sendChnCmsg(channel, {id:"WRIN", bMUTE:(on ? 0 /*mute*/: 1 /*normal*/)});
}

// ------------------------- * Channel ON/OFF processing logic * -------------------------

/*
** chOnDown
**
** called on Channel's ON key down
*/
function chOnDown(channel, local)
{
	var chn=ch_data[channel];

// Send key to phone
	if(chn.phone) onPhoneFaderEvent(channel, 'take');
	if(local)
	{
		lwcpSendCHevent(chn, {ON_But: "Down", ON_State: (chn.onsw?"On":"Off")});
	// Mute
		if(chn.onsw && chn.srct==srct.crhost)
			chnSetMute(channel, true);
	}
// Turn channel ON if allowed
	if(chnPreON(channel))
		chnSetOnOff(channel, true);
}

/*
** chOnUp
**
** called on Channel's ON key up
*/
function chOnUp(channel, local)
{
	var chn=ch_data[channel];

	if(local)
	{
		lwcpSendCHevent(chn, {ON_But: "Up", ON_State: (chn.onsw?"On":"Off")});
	// Mute
		if(chn.onsw && chn.srct==srct.crhost)
			chnSetMute(channel, false);
	}
}

/*
** chOffDown
**
** called on Channel's OFF key down
*/
function chOffDown(channel, local)
{
	var chn=ch_data[channel];

	if(local)
	{
		lwcpSendCHevent(chn, {OFF_But: "Down", ON_State: (chn.onsw?"On":"Off")});
	}
// Send key to phone
	if(chn.phone)
	{
		if(chn.onsw || chn.preview || chn.armed)
			onPhoneFaderEvent(channel, 'drop-lite');
		else
			onPhoneFaderEvent(channel, 'drop');
	}
// Turn channel OFF if allowed
	if(chnPreOFF(channel))
		chnSetOnOff(channel, false);
}

/*
** chOffUp
**
** called on Channel's OFF key up
*/
function chOffUp(channel, local)
{
	var chn=ch_data[channel];

	if(local)
	{
		lwcpSendCHevent(chn, {OFF_But: "Up", ON_State: (chn.onsw?"On":"Off")});
	}
}

/*
** chnLineDroped
**
** called when phone line is dropped, will turn channel and Preview OFF
*/
function chnLineDroped(channel)
{
	var chn;

	if(!(channel in ch_data)) return;
	chn=ch_data[channel];

	if(chn.onsw) chnSetOnOff(channel, false);
	if(chn.preview) chnSetPview(channel, false);
}

// ------------------------- * Fader Start * -------------------------

/*
** onChnFaderUp
**
** called when user moves fader UP from OFF position
*/
function onChnFaderUp(channel)
{
	var chn;

	if(!(channel in ch_data)) return;
	chn=ch_data[channel];

	chn.fader=true;
	lwcpSendCHevent(chn, {Fader_State: "UP"});
	if(chn.faderStart)
	{
	// Fader Start
		if(chn.faderStart==2)
		{
			if(chn.armed)
				chnSetOnOff(channel, true);
		}
		else if(!mod.getChnKey(channel, keys.ChnOFF))
			chnSetOnOff(channel, true);
	}
}

/*
** onChnFaderDown
**
** called when user moves fader DOWN to OFF position
*/
function onChnFaderDown(channel)
{
	var chn;

	if(!(channel in ch_data)) return;
	chn=ch_data[channel];

	chn.fader=false;
	lwcpSendCHevent(chn, {Fader_State: "DOWN"});
	if(chn.faderStart)
	{
		if(chn.faderStart==2 && chn.onsw) chn.armed=true;
		chnSetOnOff(channel, false);
	}
}

/*
** chnSetFaderStart
*/
function chnSetFaderStart(channel, faderStart)
{
	var chn=ch_data[channel];

	chn.faderStart=faderStart;
	if(faderStart && chn.onsw && !chn.fader)
	{
		if(faderStart==2) chn.armed=true;
		chnSetOnOff(channel, false); // Turn channel OFF
	}
	if(faderStart!=2 && chn.armed)
	{
	// disarm fader
		chn.armed=false;
		updateChannelOnOff(channel);
	}
}

// ------------------------- * AB switch * -------------------------

/*
** chnABswitch
*/
function chnABswitch(channel)
{
	var chn=ch_data[channel];
	var rmsg, src;
	var n, tag;
	var cur_srci, new_srci=null, take_next=false;

	rmsg=host.sendChnCmsgRq(channel, {id:'READ', bSRCI:0});
	if(!rmsg || !('dSRCI' in rmsg)) return;
	cur_srci=rmsg.dSRCI;

	rmsg=getSourceList(channel, true);
	if(rmsg==null || !('wSRLC' in rmsg) || !('mSRLE' in rmsg)) return;

	for(n=1; n<=rmsg.wSRLC; n++)
	{
		tag=host.sprintf("m%04d", n);
		if(!(tag in rmsg.mSRLE)) continue;
		src=rmsg.mSRLE[tag];
		if(!('dSRCI' in src)) continue;

		if(new_srci==null || take_next) new_srci=src.dSRCI;
		take_next=(cur_srci==src.dSRCI);
	}

	if(new_srci!=null && new_srci!=cur_srci)
		host.sendChnCmsg(channel, {id:'WRIN', dSRCI:new_srci});
}

// ------------------------- * event processing * -------------------------

/*
** onChnKeyDown
**
** called when user presses the key
*/
function onChnKeyDown(channel, key)
{
	var chn;
	var rmsg;

	if(!(channel in ch_data)) return;
	chn=ch_data[channel];

	switch(key)
	{
	case keys.ChnSet:
		if(chn.phone) onPhoneFaderEvent(channel, 'phone');
		break;
	case keys.ChnHold:
		if(chn.phone) onPhoneFaderEvent(channel, 'hold');
		break;
	case keys.ChnON:
		if(chn.enbl.onoff) chOnDown(channel, true);
		break;
	case keys.ChnOFF:
		if(!chn.enbl.onoff) break;
		if(chn.allowAB && !chn.onsw)
		{
			chn.allowAB=0;
			chnABswitch(channel);
			chn.allowAB=mod.getChnKey(channel, keys.ChnOptions)?1:false;
		}
		else
			chOffDown(channel, true);
		break;
	case keys.ChnPGM1:
		if(chn.enbl.pgm[1]) chToggleVal(channel, chn.pgm[1], "bPGM1");
		break;
	case keys.ChnPGM2:
		if(chn.enbl.pgm[2]) chToggleVal(channel, chn.pgm[2], "bPGM2");
		break;
	case keys.ChnPGM3:
		if(chn.enbl.pgm[3]) chToggleVal(channel, chn.pgm[3], "bPGM3");
		break;
	case keys.ChnPGM4:
		if(chn.enbl.pgm[4]) chToggleVal(channel, chn.pgm[4], "bPGM4");
		break;
	case keys.ChnPreview: // toggle channel preview, possible group preview
		if(!chn.enbl.preview) break;
		var otherkey=chnIsPviewDown();

		if(otherkey)
			ch_preview.grmode=true;
		else
		{
			ch_preview.grmode=false;
			ch_preview.grkey=channel;
		}

		if(!chn.preview) // preview was OFF before pressing
		{
			if(chn.talkback)
				chn.tbck_latch=true; // make talkback latch (will work if talkback is now pressed)
			chn.pview_key=1;
			if(chn.phone) onPhoneFaderEvent(channel, 'take');
			if(!otherkey && pview_interlock && !chn.tbck_latch) chnTurnOffAllPview();
			chnSetPview(channel, true);
		}
		else
			chn.pview_key=2;
		break;
	case keys.ChnTalkback:
		if(!chn.enbl.tbck) break;
		chn.tbck_latch=(chn.tbck_latch?false:0);
		host.sendChnCmsg(channel, {id:'WRIN', bTBCK:1});
		break;
	case keys.ChnOptions:
		if(chn.allowAB!==false) break;
		if(pages.open==pages.main)
		{
			chn.allowAB=true;
			break;
		}
		if(!chn.enbl.options) break;
		if(('channel' in pages.open) && pages.open.channel>=1 && pages.open.channel<=64 && pages.open.channel!=channel)
			chSwitchToChannel(channel); // switch to this channel
		else
			activatePage(pages.chopt, channel); // open default option page or close whether option page is opened
		break;
	}
}

/*
** onChnKeyUp
**
** called when user releases the key
*/
function onChnKeyUp(channel, key)
{
	var chn;

	if(!(channel in ch_data)) return;
	chn=ch_data[channel];

	switch(key)
	{
	case keys.ChnON:
		if(chn.enbl.onoff) chOnUp(channel, true);
		break;
	case keys.ChnOFF:
		if(chn.enbl.onoff) chOffUp(channel, true);
		break;
	case keys.ChnPreview:
		if(!chn.enbl.preview) break;
		if(chn.pview_key==0) break;

		var prev_state, otherkey;

		prev_state=(chn.pview_key==2);
		chn.pview_key=0;
		otherkey=chnIsPviewDown();
		if(ch_preview.grmode)
		{
			if(!otherkey)
				ch_preview.grmode=false;
			if(ch_preview.grkey!=channel && prev_state)
				chnSetPview(channel, false);
		}
		else
		{
			if(prev_state)
			{
			// if PREV state on key down event was ON
				if(!pview_interlock || chn.talkback)
					chnSetPview(channel, false); // turn off only current channel pview
				else
					chnTurnOffAllPview(); // turn preview OFF for all the rest channels that has preview ON
				if(chn.phone) onPhoneFaderEvent(channel, 'drop-lite');
			}
		}
		break;
	case keys.ChnOptions:
		if(chn.enbl.options)
		{
			if(chn.allowAB===true)
			{
				chn.allowAB=false;
				activatePage(pages.chopt, channel); // open default option page or close whether option page is opened
			}
		}
		chn.allowAB=false;
		break;
	case keys.ChnTalkback:
		if(!chn.enbl.tbck) break;
		if(!chn.tbck_latch)
		{
			host.sendChnCmsg(channel, {id:'WRIN', bTBCK:0});
			if(chn.tbck_latch===false)
			{
				if(chn.preview && chn.pview_key==0)
					chnSetPview(channel, false);
			}
			else
				chn.tbck_latch=false;
		}
		break;
	}
}

/*
** onChnRotaryEvent
**
** called when user turns rotary knob
*/
function onChnRotaryEvent(channel, delta, newValue)
{
	var chn=ch_data[channel];

	switch(chn.knob_fn)
	{
	case knobFn.none:
	default:
		return;
	case knobFn.fatg:
		//if(chn.gain_lock) return;
		host.sendChnCmsg(channel, {id:"WRIN", wFATG: valAdjust(chn.gain, delta, {min: -250, max: 250, step: 5, units: "dB"})});
		break;
	case knobFn.gain:
		if(!chn.src_lwrp) return;
		host.lwctrlSetGain(chn.lpid, valAdjust(chn.src_gain, delta, {min: -200, max: 850, step: 5, units: "dB"}));
		break;
	case knobFn.pano:
	// TODO!
		return;
	case knobFn.auxA:
		//host.sendChnCmsg(channel, {id:"WRIN", wA1GN: valAdjust(chn.aux_gain[1], delta, aux_const)});
		break;
	case knobFn.auxB:
		//host.sendChnCmsg(channel, {id:"WRIN", wA2GN: valAdjust(chn.aux_gain[2], delta, aux_const)});
		break;
	case knobFn.auxC:
		//host.sendChnCmsg(channel, {id:"WRIN", wA3GN: valAdjust(chn.aux_gain[3], delta, aux_const)});
		break;
	case knobFn.auxD:
		//host.sendChnCmsg(channel, {id:"WRIN", wA4GN: valAdjust(chn.aux_gain[4], delta, aux_const)});
		break;
	}
	chn.knob_to=host.resetTimeout(chn.knob_to, 1000, function(){this.knob_to=null;updateChnDisp(this.id);}, chn);
}

/*
** chnInitPhones
**
** function is called when we got some info about phone channel from ACL
** bind channel to US or EU phone system
** US phone parameters:
**  bPHCH - connection ID (phone_id) for US phone module faders (0 - none)
**  bHYNR - hybrid number for US phone module faders (0 - none)
**  bCALW - conferencing allowed (AKA mashing) for US phone module faders (0 - not)
**  bFXLN - fixed line for US phone module faders (0 - none)
** EU phone parameters:
**  sPHIP - phone box IP
**  dPHLN - phone line:
**    >0 => phone type is Telos2, line=dPHLN
**    =0 => phone type is NONE
**    <0 => phone type is Nx12 or compatible, line=(-dPHLN)&63, hybrid=((-dPHLN)>>6)&15, split=((-dPHN)&1024)?true:false;
*/
function chnInitPhones(channel, chmsg)
{
	var chn=ch_data[channel], n;

// select a different line, if current is disconnected
	if(last_used_phone_ch==channel)
	{
		last_used_phone_ch=0; // in case there is no active lines
		for(n=1; n<=64; n++)
			if(ch_data[n].phone)
		{
			last_used_phone_ch=n; // pick the first of EU or US phone channels
			break;
		}
	}
// clear any phone state we might have
	chn.phone=false;
	phUS_UpdateFaderData(channel, 0, 0, 0, false);
	phEU_UpdateFaderData(channel, '', 0, 0);
	setChnPhoneIcon(channel, 0);
	setChnPhoneText(channel, "");
	setChnPhoneOnAir(channel, false);
	setChnPhoneLed(channel, 0);

// --- test if we are assigned as a US phone hybrid

// All these have to arrive together, otherwise they will cause a fader clear!!!
	if(('bPHCN' in chmsg) && chmsg.bPHCN!=0
		&& ('bHYNR' in chmsg) && chmsg.bHYNR!=0
		&& ('bCALW' in chmsg)
		&& ('bFXLN' in chmsg))
	{
		chn.phone=phUS_UpdateFaderData(channel, chmsg.bPHCN, chmsg.bHYNR, chmsg.bFXLN, chmsg.bCALW!=0);
		return;
	}
// --- check for EU connection
	if(('sPHIP' in chmsg) && ('dPHLN' in chmsg))
	{
		var addr=chmsg.sPHIP;
		var line=chmsg.dPHLN;

		if(addr!="" && line!=0)
		{
			log.info_f("DBG: PhoneEU: (chn%02d) got PHIP: %s/%d", channel, addr, line);
			if(line>0)
			{
				if(addr.substr(0, 3)=="vx:")
				{
				/* TelosVX */
					chn.phone=phEU_UpdateFaderData(channel, addr, line, 0);
				}
				else
				{
				/* Telos 2 */
					chn.phone=phEU_UpdateFaderData(channel, 't2:'+addr+'/'+line, 1, 0);
				}
			}
			else if(line<0)
			{
			/* Nx12 or compatible, hybrid = ((-line)>>6)&15, line = ((-line)&63)-1, shskip = ((-line)&1024)?1:0 */
				line=-line;
				chn.phone=phEU_UpdateFaderData(channel, 'ap:'+addr+((line&1024)?"/1":''), ((line>>6)&15)+1, (line&63)-1);
			}
		// select line, if none is selected yet
			if(!last_used_phone_ch) last_used_phone_ch=channel;
		}
	}
}

/*
** chnPGMindi
*/
function chnPGMindi(channel, pgm, value)
{
	var chn=ch_data[channel];
	var led_style;

	if(value==0)
	{
		chn.pgm[pgm]=0;
		led_style=0;
		if(!chnIsPgmOn(pgm)) mtr_label[pgm].activate(false);
	}
	else
	{
		chn.pgm[pgm]=1;
		led_style=(pgm==4 && recmode_enabled && recmode_active)?4:255;
		mtr_label[pgm].activate(true);
	}
	mod.setChnKeyLed(channel, keys.ChnPGM1+(pgm-1), led_style);
	ch_pgmx_event(chn, pgm);
}

/*
** onChnRecvMsg
**
** called on incoming Channel message
** {NEST,mCHxx=<...>}
*/
function onChnRecvMsg(channel, chmsg)
{
	var chn;
	var i;

	if(!(channel in ch_data)) return;
	chn=ch_data[channel];

	if(chmsg.id!='INDI') return;

// Source sharing status
	if('sSCON' in chmsg) chn.src_owner_name =chmsg.sSCON;
	if('bSCFN' in chmsg) chn.src_owner_fader=chmsg.bSCFN;

// Channel source name /* FIXME: there is a bug in ACL, it sends NXTN without SRCN */
	if(('bvSRCN' in chmsg) || ('bvNXTN' in chmsg))
	{
		var srcn, nxtn;

		srcn=('bvSRCN' in chmsg)?host.arrtostr(chmsg.bvSRCN):chn.src_name;
		nxtn=('bvNXTN' in chmsg)?host.arrtostr(chmsg.bvNXTN):"";

		chn.src_name=srcn;
		chn.src_next=nxtn;
		updateChnDisp(channel);
	}
	if('sLABL' in chmsg)
	{
		chn.src_label=chmsg.sLABL;
		updateChnDisp(channel);
	}
	if('bDSPL' in chmsg)
	{
		chn.lbl_dspl=chmsg.bDSPL;
		updateChnDisp(channel);
	}
	if('dSRCI' in chmsg)
	{
		chn.srci=chmsg.dSRCI;
		lwcpSendCHevent(chn, {src_id: chn.srci, src_name: '"'+chn.src_name+'"', src_lwch: (chn.lpid?chn.lpid:-1)});
	}

// Fader Gain
	if(('wFAGN' in chmsg) && chmsg.wFAGN!=0x7FFF)
	{
		mod.setChnFader(channel, chmsg.wFAGN/10.0);
	}
	if('wFATG' in chmsg)
	{
		chn.gain=chmsg.wFATG;
		if(chn.knob_to!=null && chn.knob_fn==knobFn.fatg)
			mod.setChnText(channel, host.sprintf("%+5.1f dB", chn.gain/10), mod.CENTER);
	}
	//if('bFATL' in chmsg)
	//	chn.gain_lock=(chmsg.bFATL!=1);

// Channel lock mask
	if('wENBL' in chmsg)
	{
		chn.enbl.pgm[1] =(chmsg.wENBL&(1<<0))!=0;
		chn.enbl.pgm[2] =(chmsg.wENBL&(1<<1))!=0;
		chn.enbl.pgm[3] =(chmsg.wENBL&(1<<2))!=0;
		chn.enbl.pgm[4] =(chmsg.wENBL&(1<<3))!=0;
		chn.enbl.options=(chmsg.wENBL&(1<<4))!=0;
		chn.enbl.preview=(chmsg.wENBL&(1<<5))!=0;
		chn.enbl.onoff  =(chmsg.wENBL&((1<<6)|(1<<7)))!=0;
		chn.enbl.fader  =(chmsg.wENBL&(1<<8))!=0;
		chn.enbl.tbck   =(chmsg.wENBL&(1<<9))!=0;
		chn.enbl.hpsrc  =(chmsg.wENBL&(1<<10))!=0;
	// apply fader filter
		host.setFaderFilter(channel, chn.enbl.fader);
	}
// Channel preview mode
	if('bPVMO' in chmsg) chn.pvmo=chmsg.bPVMO;
// Channel source type
	if('bSRCT' in chmsg) chn.srct=chmsg.bSRCT;
// Channel source status
	if('bSRCS' in chmsg)
	{
		chn.src_status=chmsg.bSRCS;
		updateChnDisp(channel);
		updateChnIcon(channel);
	}
// Group start
	if('bGRPS' in chmsg)
	{
		switch(chmsg.bGRPS)
		{
		case 3: // Master
			chn.groupStart=1;
			break;
		case 1: // Slave
			chn.groupStart=2;
			break;
		default: // Otherwise Independent
			chn.groupStart=false;
			break;
		}
	}
// Fader Start
	if('bFAMO' in chmsg) chnSetFaderStart(channel, chmsg.bFAMO);
// Ready
	if('bRDEN' in chmsg) chn.rden=chmsg.bRDEN;
	if('bRDCM' in chmsg) chn.rdcm=chmsg.bRDCM;
// Channel ON/OFF lamps
	if(('bONSW' in chmsg) && (chmsg.bONSW!=0)!=chn.onsw)
	{
		if(chmsg.bONSW==0)
		{
			chn.onsw=false;
			if(chn.fader) chn.armed=false; // external OFF disarms fader
			if(chn.phone) onPhoneFaderEvent(channel, 'ch-off'); // notify phone system that fader went off
		}
		else
		{
			chn.onsw=true;
			if((chn.pvmo==2 || chn.pvmo==4) && chn.preview)
				chnSetPview(channel, false);
		}
		if(chn.lpid!=0 && (chn.srct==srct.crguest || chn.srct==srct.stguest || chn.srct==srct.extmic))
			sendLwCHmsg(chn.lpid, 'SET', 'LMP_ON='+((chn.onsw && !chn.mute)?'ON':'OFF')+',LMP_OFF='+((!chn.onsw)?'ON':'OFF'));
		lwcpSendCHevent(chn, {ON_State: (chn.onsw?"ON":"OFF")});
	}
	if(('bRDEN' in chmsg) || ('bRDCM' in chmsg) || ('bONSW' in chmsg)) updateChannelOnOff(channel);
// PGMx lamps
	if('bPGM1' in chmsg) chnPGMindi(channel, 1, chmsg.bPGM1);
	if('bPGM2' in chmsg) chnPGMindi(channel, 2, chmsg.bPGM2);
	if('bPGM3' in chmsg) chnPGMindi(channel, 3, chmsg.bPGM3);
	if('bPGM4' in chmsg) chnPGMindi(channel, 4, chmsg.bPGM4);
// PREV lamps
	if('bPREV' in chmsg)
	{
		if(chmsg.bPREV!=0)
		{
			chn.preview=true;
			mod.setChnKeyLed(channel, keys.ChnPreview, 255);
			if((chn.pvmo==3 || chn.pvmo==4) && chn.onsw)
				chnSetOnOff(channel, false);
		}
		else
		{
		// if preview went off send 'OFF key' to EU phone
			if(chn.preview==true && chn.phone) onPhoneFaderEvent(channel, 'pv-off'); // notify phone system that preview went off
			chn.preview=false;
			mod.setChnKeyLed(channel, keys.ChnPreview, 0);
		}
		lwcpSendCHevent(chn, {'Asg_PREV': (chn.preview?"ON":"OFF")});
	}
// TBCK (talkback) lamps
	if('bTBCK' in chmsg)
	{
		chn.talkback=(chmsg.bTBCK==1);
		mod.setChnKeyLed(channel, keys.ChnTalkback, chn.talkback?255:0);
	}
// MUTE state
	if('bMUTE' in chmsg)
	{
		chn.mute=(chmsg.bMUTE==0); // 0 - muted; 1 - not muted
		updateChannelOnOff(channel);
		if(chn.lpid!=0 && (chn.srct==srct.crguest || chn.srct==srct.stguest || chn.srct==srct.extmic))
			sendLwCHmsg(chn.lpid, 'SET', 'LMP_MUTE='+(chn.mute?'ON':'OFF')+(chn.onsw?',LMP_ON='+(chn.mute?'OFF':'ON'):''));
		lwcpSendCHevent(chn, {Mute_State: (chn.mute?"ON":"OFF")});
	}
// TTCR (talk to cr)
	if('bTTCR' in chmsg)
	{
		chn.talkCR=(chmsg.bTTCR==1);
		if(chn.lpid!=0 && (chn.srct==srct.crguest || chn.srct==srct.stguest || chn.srct==srct.extmic))
			sendLwCHmsg(chn.lpid, 'SET', 'LMP_TALK='+(chn.talkCR?'ON':'OFF'));
	}
// SDSP - indicate icon
	if('bSDSP' in chmsg)
	{
		chn.dplay_stat=chmsg.bSDSP;
		updateChnIcon(channel);
	}
	if(('sPHIP' in chmsg) || ('bPHCN' in chmsg))
		chnInitPhones(channel, chmsg);
// ID for remote control. If remote control enabled, the same as primary source id (PSID), else 0
	if(('dLPID' in chmsg) && chn.lpid!=chmsg.dLPID)
	{
		if(chn.lpid!=0)
		{
			switch_off_acc_module(chn);
			if(chn.src_lwrp)
			{
				host.lwctrlDropSrc(chn.lpid);
				chn.src_lwrp=false;
			}
		}
		chn.lpid=chmsg.dLPID;
		if(chn.lpid && chn.knob_fn==knobFn.gain) chn.src_lwrp=host.lwctrlAddSrc(chn.lpid);
	}
	if(('bPRIM' in chmsg) && chn.knob_fn!=chmsg.bPRIM)
	{
		if(chn.knob_fn==knobFn.gain && chn.src_lwrp)
		{
			host.lwctrlDropSrc(chn.lpid);
			chn.src_lwrp=false;
		}
		chn.knob_fn=chmsg.bPRIM;
		switch(chn.knob_fn)
		{
		case knobFn.gain:
			if(chn.lpid) chn.src_lwrp=host.lwctrlAddSrc(chn.lpid);
			break;
		case knobFn.fatg:
			host.sendChnCmsg(channel, {id:'READ', bFATG:1, bFATL:1});
			break;
		case knobFn.pano:
			host.sendChnCmsg(channel, {id:'READ', bPANO:1});
			return;
		case knobFn.auxA:
			host.sendChnCmsg(channel, {id:'READ', bA1GN:1});
			break;
		case knobFn.auxB:
			host.sendChnCmsg(channel, {id:'READ', bA2GN:1});
			break;
		case knobFn.auxC:
			host.sendChnCmsg(channel, {id:'READ', bA3GN:1});
			break;
		case knobFn.auxD:
			host.sendChnCmsg(channel, {id:'READ', bA4GN:1});
			break;
		}
	}
// indication where HP selector is
	if('bvHPTG' in chmsg)
	{
		chn.hptg=host.arrtostr(chmsg.bvHPTG);
		chn.hptg_scroll=chn.hptg;
		if(chn.lpid!=0 && (chn.srct==srct.crguest || chn.srct==srct.stguest || chn.srct==srct.extmic))
		{
		// turn on preset lamp if corresponding tag
			sendLwCHmsg(chn.lpid, 'SET',
				'LMP_HPpset1='+(chn.hptg==chn.hp_pset1?'ON':'OFF')+
				',LMP_HPpset2='+(chn.hptg==chn.hp_pset2?'ON':'OFF')+
				(chn.onsw?',LMP_ON=ON,LMP_OFF=OFF':',LMP_ON=OFF,LMP_OFF=ON'));
		}
	}
// indication HP preset 1
	if('mHPP1' in chmsg)
	{
		if(('bvHPTG' in chmsg.mHPP1) && chn.lpid!=0)
			chn.hp_pset1=host.arrtostr(chmsg.mHPP1.bvHPTG);
	}
// indication HP preset 2
	if('mHPP2' in chmsg)
	{
		if(('bvHPTG' in chmsg.mHPP2) && chn.lpid!=0)
			chn.hp_pset2=host.arrtostr(chmsg.mHPP2.bvHPTG);
	}
// indication HP current source name
	if('bvHPSN' in chmsg)
	{
		if(chn.lpid!=0 && (chn.srct==srct.crguest || chn.srct==srct.stguest || chn.srct==srct.extmic))
		{
			chn.hp_src_name=host.arrtostr(chmsg.bvHPSN);
			sendLwCHmsg(chn.lpid, 'SET', 'DSP_HPtext="'+chn.hp_src_name+'"');
		}
	}
// indication HP gain
	if('wHPGN' in chmsg)
	{
		if(chn.lpid!=0)
			chn.hp_gain=mon_vol_db_to_idx[clamp(chmsg.wHPGN, mon_const.gain_min, mon_const.gain_max)];
	}

// Open page - gives message to processing by page message handlers
	var handl=pages.rxHandlersCh[channel];
	for(i in handl)
	{
		if('onIndi' in handl[i]) handl[i].onIndi(chmsg);
	}
}

/*
** retPGMindi
*/
function retPGMindi(channel, pgm, value)
{
	var ret=ret_data[channel];

	if(value==0)
	{
		ret.pgm[pgm]=0;
		if(!chnIsPgmOn(pgm)) mtr_label[pgm].activate(false);
	}
	else
	{
		ret.pgm[pgm]=1;
		mtr_label[pgm].activate(true);
	}
}

/*
** onRetRecvMsg
**
** called on incoming Return message
** {NEST,mRETx=<...>}
*/
function onRetRecvMsg(channel, retmsg)
{
	var ret;

	if(!(channel in ret_data)) return;
	ret=ret_data[channel];

	if('bONSW' in retmsg) ret.onsw=(retmsg.bONSW!=0);
	if('bPGM1' in retmsg) retPGMindi(channel, 1, retmsg.bPGM1);
	if('bPGM2' in retmsg) retPGMindi(channel, 2, retmsg.bPGM2);
	if('bPGM3' in retmsg) retPGMindi(channel, 3, retmsg.bPGM3);
	if('bPGM4' in retmsg) retPGMindi(channel, 4, retmsg.bPGM4);
}

/*
** onLWgainChange
*/
function onLWgainChange(psid, gain)
{
	var n;

	for(n=1; n<=64; n++)
	{
		if(ch_data[n].exists && ch_data[n].lpid==psid && ch_data[n].src_lwrp)
		{
			ch_data[n].src_gain=gain;
			if(ch_data[n].knob_to!=null && ch_data[n].knob_fn==knobFn.gain)
				mod.setChnText(n, host.sprintf("%+5.1f dB", ch_data[n].src_gain/10), mod.CENTER);
			if(('channel' in pages.open) && pages.open.channel==n && ('onGainChange' in pages.open))
				pages.open.onGainChange(gain);
		}
	}
//	log.info_f("gain of %d is %d", psid, gain);
}

// ------------------------- * Remote control * -------------------------

/*
** onChnGPIOevent
**
** called on incoming GPIO event
*/
function onChnGPIOevent(channel, circuit, value)
{
	var chn=ch_data[channel];

// ON and off are common to all source types
	if(circuit==gpioCirc.on)
	{
		if(value==1) chOnDown(channel, false);
		return;
	}
	if(circuit==gpioCirc.off)
	{
		if(value==1) chOffDown(channel, false);
		return;
	}

	switch(chn.srct)
	{
	case srct.crhost:
	case srct.producer:
		switch(circuit)
		{
		case gpioCirc.talkToST:
			host.sendChnCmsg(channel, {id:"WRIN", bTTST:value});
			break;
		case gpioCirc.mute:
			chnSetMute(channel, value);
			break;
		case gpioCirc.talkToPreview:
			host.sendChnCmsg(channel, {id:"WRIN", bTTPV:value});
			break;
		}
		break;
	case srct.crguest:
	case srct.stguest:
	case srct.extmic:
	case srct.codec:
		switch(circuit)
		{
		case gpioCirc.talkToCR:
			host.sendChnCmsg(channel, {id:"WRIN", bTTCR:value});
			break;
		case gpioCirc.mute:
			chnSetMute(channel, value);
			break;
		case gpioCirc.talkToSource:
			host.sendChnCmsg(channel, {id:"WRIN", bTBCK:value});
			break;
		}
		break;
	case srct.line:
	case srct.phone:
	case srct.computer:
		switch(circuit)
		{
		case gpioCirc.preview:
			host.sendChnCmsg(channel, {id:"WRIN", bPREV:value});
			//xxx: chnSetPview(channel, value);
			break;
		case gpioCirc.reset:
			if(chn.srct==srct.computer) break; // Computer has no reset
			if(chn.onsw && value==1)
			{
			// send reset track command to acl: generally this is OFF, but without OFF pulse
				host.sendChnCmsg(channel, {id:"WRIN", bCMND:7});
			}
			break;
		case gpioCirc.ready:
			host.sendChnCmsg(channel, {id:"WRIN", bRDCM:value});
			break;
		}
		break;
	}
}

/*
** onChnLWCPevent
**
** called on incoming LWCP message: <op> LwCH#<lpid> {props}
** message can be multicast: connId==LWCP_sidMultiCast or not (which IMO doesn't make much sense)
*/
function onChnLWCPevent(channel, connId, op, props)
{
	var chn=ch_data[channel];
	var ctag, string_tag;
	var rmsg;

	if(op=='SET')
	{
		if('Mute_State' in props) // values: MUTED | NORMAL
		{
			if(props.Mute_State=='MUTED')
				chnSetMute(channel, true);
			else if(props.Mute_State=='NORMAL')
				chnSetMute(channel, false);
		}
		if('ON_State' in props) // values: ON | OFF
		{
			if(props.ON_State=='ON')
				chnSetOnOff(channel, true);
			else if(props.ON_State=='OFF')
				chnSetOnOff(channel, false);
		}
		if('Fader_Gain' in props)
		{
			mod.setChnFader(channel, props.Fader_Gain);
		}
		if('Asg_PGM1' in props)
		{
			if(props.Asg_PGM1=='ON')
				host.sendChnCmsg(channel, {id:'WRIN', bPGM1:1});
			else if(props.Asg_PGM1=='OFF')
				host.sendChnCmsg(channel, {id:'WRIN', bPGM1:0});
		}
		if('Asg_PGM2' in props)
		{
			if(props.Asg_PGM2=='ON')
				host.sendChnCmsg(channel, {id:'WRIN', bPGM2:1});
			else if(props.Asg_PGM2=='OFF')
				host.sendChnCmsg(channel, {id:'WRIN', bPGM2:0});
		}
		if('Asg_PGM3' in props)
		{
			if(props.Asg_PGM3=='ON')
				host.sendChnCmsg(channel, {id:'WRIN', bPGM3:1});
			else if(props.Asg_PGM3=='OFF')
				host.sendChnCmsg(channel, {id:'WRIN', bPGM3:0});
		}
		if('Asg_PGM4' in props)
		{
			if(props.Asg_PGM4=='ON')
				host.sendChnCmsg(channel, {id:'WRIN', bPGM4:1});
			else if(props.Asg_PGM4=='OFF')
				host.sendChnCmsg(channel, {id:'WRIN', bPGM4:0});
		}
		if('Asg_PREV' in props)
		{
			if(props.Asg_PREV=='ON')
				host.sendChnCmsg(channel, {id:'WRIN', bPREV:1});
			else if(props.Asg_PREV=='OFF')
				host.sendChnCmsg(channel, {id:'WRIN', bPREV:0});
		}
		if(mstEnable.chn_src)
		{
		/* source change is allowed */
			if('src_id'   in props) host.sendChnCmsg(channel, {id:'WRIN', dSRCI:props.src_id});
			if('icom_src' in props) host.sendChnCmsg(channel, {id:'WRIN', dPSID:props.icom_src});
		}
	}
	else if(op=='GET')
	{
		rmsg={};

		if('Fader_Gain' in props) rmsg.Fader_Gain=Math.round(mod.getChnFader(chn.id)*10)/10;
		if('Fader_State' in props) rmsg.Fader_State=(chn.fader?"UP":"DOWN");
		if('Mute_State' in props) rmsg.Mute_State=(chn.mute?"ON":"OFF");
		if('ON_State' in props) rmsg.ON_State=(chn.onsw?"ON":"OFF");
		if('Asg_PGM1' in props) rmsg.Asg_PGM1=(chn.pgm[1]?"ON":"OFF");
		if('Asg_PGM2' in props) rmsg.Asg_PGM2=(chn.pgm[2]?"ON":"OFF");
		if('Asg_PGM3' in props) rmsg.Asg_PGM3=(chn.pgm[3]?"ON":"OFF");
		if('Asg_PGM4' in props) rmsg.Asg_PGM4=(chn.pgm[4]?"ON":"OFF");
		if('Asg_PREV' in props) rmsg.Asg_PREV=(chn.preview?"ON":"OFF");
		if('src_lwch' in props) rmsg.src_lwch=(chn.lpid?chn.lpid:-1);
		if('src_id' in props) rmsg.src_id=chn.srci;
		if('src_name' in props) rmsg.src_name='"'+chn.src_name+'"';
		if('src_list' in props) rmsg.src_list=lwGetSrcProfList(chn.id);
		return rmsg;
	}
	else if(op=='EVENT' && connId==LWCP_sidMultiCast && (chn.srct==srct.crguest || chn.srct==srct.stguest || chn.srct==srct.extmic))
	{
	// Multicast Events are processed here (now we have events only for guests)
	// this is so called Accessory Module interface
		if('BTN_ON' in props)
		{
			if(props.BTN_ON=='DOWN')
				chOnDown(channel, false);
			else
				chOnUp(channel, false);
		}
		if('BTN_OFF' in props)
		{
			if(props.BTN_OFF=='DOWN')
				chOffDown(channel, false);
			else
				chOffUp(channel, false);
		}
		if('BTN_TALK' in props)
		{
			host.sendChnCmsg(channel, {id:"WRIN", bTTCR: (props.BTN_TALK=='DOWN')?1:0});
		}
		if('BTN_MUTE' in props)
		{
			chnSetMute(channel, props.BTN_MUTE=='DOWN');
		}

		if(('BTN_HPsel' in props) && props.BTN_HPsel=='DOWN') // knob press
		{
			if(chn.presetMode>0)
				selectPreset(channel, chn.presetMode);
			else
			{
				chn.srcresetTimeout=host.clearTimeout(chn.srcresetTimeout);
				if(chn.gainconfirmTimeout!=null)
				{
					chn.gainconfirmTimeout=host.clearTimeout(chn.gainconfirmTimeout);
				// remove volume indicator and show source name: TODO: do this a better way!
					host.sendChnCmsg(chn.id, {id:'WRIN', wHPGN:mon_vol_idx_to_db[chn.hp_gain]});
					sendLwCHmsg(chn.lpid, 'SET', 'DSP_HPtext="'+chn.hp_src_name+'"');
				}
				else
				{
					if(chn.enbl.hpsrc)
					{
						chn.hp_scroll_mode=(chn.hp_scroll_mode==0)?1:0;
						if(chn.hp_scroll_mode==1)
						{
						// enable source selection mode, display current selection and start blinking
							load_src_list();
							chn.hptg_scroll=chn.hptg;
							sendLwCHmsg(chn.lpid, 'SET', 'DSP_HPtext="*'+chn.hp_src_name+'"');
							object_blinker_acc_mod.add(chn, function(x) {return x.hp_scroll_mode==1;}, true, acc_mod_on, acc_mod_off);
							chn.srcresetTimeout=host.setTimeout(4000, srcreset, chn);
						}
						else
						{
						// accept new source and stop blinking
							object_blinker_acc_mod.remove(chn);
							ctag=(chn.hptg_scroll!="")?chn.hptg_scroll:chn.hptg;
							host.sendChnCmsg(channel, {id:"WRIN", bvHPTG:strToArr(ctag)});
						}
					}
				}
			}
		}
		if('ROT_HPsel' in props) // knob rotate
		{
			if(chn.hp_scroll_mode==1 || chn.hp_scroll_mode==2)
			{
			//source selection
				ctag=(chn.hptg_scroll!="")?chn.hptg_scroll:chn.hptg;
			// get next tag
				chn.hptg_scroll=acc_scroll_tag(ctag, props.ROT_HPsel);
			// set text
				if((chn.presetMode==0 && chn.hptg_scroll==chn.hptg)      // selecting main source
				|| (chn.presetMode==1 && chn.hptg_scroll==chn.hp_pset1)  // selecting Preset 1 source
				|| (chn.presetMode==2 && chn.hptg_scroll==chn.hp_pset2)) // selecting Preset 2 source
				{
					string_tag=(chn.presetMode==0)?"*":">";
				}
				else
					string_tag="";
				sendLwCHmsg(chn.lpid, 'SET', 'DSP_HPtext="'+string_tag+acc_list_text(chn.hptg_scroll)+'"');
				chn.srcresetTimeout=host.resetTimeout(chn.srcresetTimeout, 4000, srcreset, chn);
			}
			else
			{
			// volume control
				chn.hp_gain=clamp(chn.hp_gain+props.ROT_HPsel*34, 0, 1000);
				host.sendChnCmsg(channel, {id:"WRNI", wHPGN:mon_vol_idx_to_db[chn.hp_gain]});
				indicateHPGain(channel, chn.lpid);
				chn.gainconfirmTimeout=host.resetTimeout(chn.gainconfirmTimeout, 100, confirmHPGain, chn);
			}
		}
		if('BTN_HPpset1' in props)
		{
			chn.presetTimeout=host.clearTimeout(chn.presetTimeout);
			if(chn.enbl.hpsrc)
			{
				if(props.BTN_HPpset1=='DOWN')
					selectPreset(channel, 1);
				else if(chn.hp_scroll_mode==0 && chn.presetMode==1)
				{
				// set current src to preset1
					host.sendChnCmsg(channel, {id:"WRIN", bvHPTG:strToArr(chn.hp_pset1)});
					chn.presetMode=0;
				}
			}
		}
		if('BTN_HPpset2' in props)
		{
			chn.presetTimeout=host.clearTimeout(chn.presetTimeout);
			if(chn.enbl.hpsrc)
			{
				if(props.BTN_HPpset2=='DOWN')
					selectPreset(channel, 2);
				else if(chn.hp_scroll_mode==0 && chn.presetMode==2)
				{
				// set current src to preset2
					host.sendChnCmsg(channel, {id:"WRIN", bvHPTG:strToArr(chn.hp_pset2)});
					chn.presetMode=0;
				}
			}
		}
	}
	else if(op=='UPDATE' && connId==LWCP_sidMultiCast && (chn.srct==srct.crguest || chn.srct==srct.stguest || chn.srct==srct.extmic))
	{
	// Accessory Module is requesting Update
		rmsg={};

		if('LMP_ON'  in props) rmsg.LMP_ON =(chn.onsw && !chn.mute)?'ON':'OFF';
		if('LMP_OFF' in props) rmsg.LMP_OFF=(!chn.onsw)?'ON':'OFF';
		if('LMP_TALK' in props) rmsg.LMP_TALK=(chn.talkCR)?'ON':'OFF';
		if('LMP_MUTE' in props) rmsg.LMP_MUTE=(chn.mute)?'ON':'OFF';
		if('LMP_HPpset1' in props) rmsg.LMP_HPpset1=(chn.hptg==chn.hp_pset1)?'ON':'OFF';
		if('LMP_HPpset2' in props) rmsg.LMP_HPpset2=(chn.hptg==chn.hp_pset2)?'ON':'OFF';
		if('DSP_HPtext'  in props) rmsg.DSP_HPtext ='"'+chn.hp_src_name+'"';
		sendLwCHmsg(chn.lpid, 'SET', lwcpEncodeProps(rmsg));
	}
	return null;
}
