// backup of pre-record-mode values, used to restore state on deactivation
var rec_backup=
{
	CHxx: [],
	CRMO: null,
	CRHP: null,
	STMO: null,
	MET4: null
};

// Notes on JS closures and impact on threading:
//
// In our particular (threaded) JavaScript interpreter, it is important that if a function B
// 1) is called by a JavaScript function A (unlike, e.g., onRecvMsg() that the host invokes directly),
// 2) and it (or the functions it calls) creates a closure (on-the-spot created function that refers
//    back to the variables outside its own text)
// then it must not perform an operation that sleeps, such as sleep() or sendCmsgRq().
// Not complying with this rule results in the context created for the closure not being available
// to the host, which causes the sleep call to return not where it should in function B, but rather
// to function A right after the call to B. This typically leads to premature thread exits.
// Therefore, using the following line would have caused trouble and an explicit function was introduced.
// object_blinker.add(pages.main.hp_src, function(x) { return recmode_active; }, false, null, null);

// ------------------------- * helper functions * -------------------------

// determine whether things should keep blinking in celebration of the Record Mode
function recmode_isactive(x){return recmode_active;}

// ------------------------- * Activation * -------------------------

// start Record Mode activation (version 4) if it is enabled
function activateRecordMode()
{
	if(!recmode_enabled)
	{
		log.info("Record Mode activation disabled in current profile");
		return;
	}

// Tell ACL that Record Mode is active, we'll finish activating it later
	host.sendNestCmsg('mRCMO', {id:'WRIN', bSTAT:1});
}

function cancelRecordMode(reason)
{
	log.info(reason);
	host.sendNestCmsg('mRCMO', {id:'WRIN', bSTAT:0});
}

// finish activating the Record Mode (version 4) if all goes well (done on INDI RCMO.STAT=1)
function finishActivatingRecordMode()
{
	if(recmode_active || !recmode_enabled) return; // if active already or not enabled, bail

	var rmsg, rmsgRCMO, CHxxData, i, pgm, tag;

// Read the so-called control states and remember them...
	rec_backup.CHxx=[];
	CHxxData=[];
	for(i=1; i<=64; i++)
	{
		if(!ch_data[i].exists) continue;
		rmsg=host.sendChnCmsgRq(i, {id:'READ', bPGM1:0, bPGM2:0, bPGM3:0, bPGM4:0, bRCMO:0});

	// do sanity checks for this response
		if(rmsg==null || !('bPGM1' in rmsg) || !('bPGM2' in rmsg) || !('bPGM3' in rmsg) || !('bPGM4' in rmsg) || !('mRCMO' in rmsg))
		{
			cancelRecordMode(host.sprintf("Cannot obtain CH%02d.{PGMx,RCMO}, Record Mode activation cancelled", i));
			return;
		}

		rmsgRCMO=rmsg.mRCMO;
		for(pgm=1; pgm<=4; pgm++)
		{
			tag="bPGM"+pgm;
			if((tag in rmsgRCMO) && rmsgRCMO[tag]!=0xFF)
			{
				if(!(i in CHxxData)) CHxxData[i]={id:'WRIN'};
				if(!(i in rec_backup.CHxx)) rec_backup.CHxx[i]={id:'WRIN'};

				CHxxData[i][tag]=rmsgRCMO[tag];
				rec_backup.CHxx[i][tag]=rmsg[tag];
			}
		}
	// ON/OFF lock
		ch_data[i].recOnswLock=(('bONSW' in rmsgRCMO) && rmsgRCMO.bONSW==1)
	}

// now get the source and the like info for CRMO, CRHP and STMO because we'll have to restore them
// only HP will change mode, so store mode for HP as well
	rec_backup.CRMO=host.sendNestCmsgRq('mCRMO', {id:'READ', bSSEL:0});
	rec_backup.CRHP=host.sendNestCmsgRq('mCRHP', {id:'READ', bSSEL:0, bMODE:0, bPVMO:0});
//	rec_backup.STMO=host.sendNestCmsgRq('mSTMO', {id:'READ', bSSEL:0}); // TODO: Mon2 don't fully support SSEL
	rec_backup.STMO=host.sendNestCmsgRq('mSTMO', {id:'READ', bSTAG:0});
	rec_backup.MET4=host.sendNestCmsgRq(cfg_surroundMode?'mMET2':'mMET4', {id:'READ', bSRSW:0});

// quick sanity checks for these responses
	if(rec_backup.CRMO==null || !('bSSEL' in rec_backup.CRMO))
	{
		cancelRecordMode("Cannot obtain CRMO.SSEL, Record Mode activation cancelled");
		return;
	}
	if(rec_backup.CRHP==null || !('bSSEL' in rec_backup.CRHP) || !('bMODE' in rec_backup.CRHP) || !('bPVMO' in rec_backup.CRHP))
	{
		cancelRecordMode("Cannot obtain CRHP.{SSEL,MODE,PVMO}, Record Mode activation cancelled");
		return;
	}
//	if(rec_backup.STMO==null || !('bSSEL' in rec_backup.STMO)) // TODO: Mon2 don't fully support SSEL
	if(rec_backup.STMO==null || !('bvSTAG' in rec_backup.STMO))
	{
		cancelRecordMode("Cannot obtain STMO.STAG, Record Mode activation cancelled");
		return;
	}
	if(rec_backup.MET4==null || !('bSRSW' in rec_backup.MET4))
	{
		cancelRecordMode("Cannot obtain MET4.SRSW, Record Mode activation cancelled");
		return;
	}

// done backing-up, activate Record mode
	recmode_active=true;
	for(i in CHxxData)
	{
		if(!(i in ch_data) || !ch_data[i].exists) continue;
	// set CHxx params
		host.sendChnCmsg(i, CHxxData[i]);
	}

// switch CR headphones and monitors to "Record Mix"
// turn CR headphones to mono mode, switch OFF preview in HP mode
	host.sendNestCmsg('mCRHP', {id:'WRIN', bSSEL:monInput.rec, bPVMO:1, bMODE:mode_type.sum});
	host.sendNestCmsg('mCRMO', {id:'WRIN', bSSEL:monInput.rec});
//	host.sendNestCmsg('mSTMO', {id:'WRIN', bSSEL:monInput.rec}); // TODO: Mon2 don't fully support SSEL
	host.sendNestCmsg('mSTMO', {id:'WRIN', bvSTAG:strToArr("REC1")});

// Switch the rightmost meter to Record Mix
	host.sendNestCmsg(cfg_surroundMode?'mMET2':'mMET4', {id:'WRIN', bSRSW:3}); // 3 - "Record"

// now send the required GPIO commands to Recorder if it is defined as non-0
	if(recmode_rcdrlpid!=0)
	{
	// FIXME? Make the message format correspond to the spec?
		host.sendNestCmsg('mGPIO',
		{
			id:'WRNI',
			dLPID:recmode_rcdrlpid,
			bvPINN:[gpio.pin_start_pulse, gpio.pin_on_lamp, gpio.pin_off_lamp],
			bvVALL:[gpio.pulse, gpio.on, gpio.off]
		});
	}

// --- visual indication ---
	for(i=1; i<=64; i++)
	{
		if(!ch_data[i].exists) continue;
		rmsg=host.sendChnCmsgRq(i, {id:'READ', bREC1:0});
		if(rmsg && ('bREC1' in rmsg) && rmsg.bREC1)
			mod.setChnKeyLed(i, keys.ChnPGM4, 4); // start flashing the Record button lamp
	}

// start blinking, if we chane any parameters
	if(rec_backup.CRHP.bSSEL!=monInput.rec) object_blinker.add(pages.main.hp_src  , recmode_isactive, true, null, null);
	if(rec_backup.CRMO.bSSEL!=monInput.rec) object_blinker.add(pages.main.mon1_src, recmode_isactive, true, null, null);
//	if(rec_backup.STMO.bSSEL!=monInput.rec) object_blinker.add(pages.main.mon2_src, recmode_isactive, true, null, null); // TODO: Mon2 don't fully support SSEL
	if(host.arrtostr(rec_backup.STMO.bvSTAG)!='REC1') object_blinker.add(pages.main.mon2_src, recmode_isactive, true, null, null); // TODO: Mon2 don't fully support SSEL
	object_blinker.add(mtr_label_right.lbl, recmode_isactive, true, null, null);

// blink record mode keys
	mod.setMasterKeyLed(keys.MstRecord, 4);
	mod.setMasterKeyLed(keys.CmbRecord, 4);
	log.info("Record Mode in now ON");
}

// ------------------------- * Deactivation * -------------------------

// start deactivating the Record Mode, finish later
function deactivateRecordMode()
{
// Tell ACL the Record Mode is off
	host.sendNestCmsg('mRCMO', {id:'WRIN', bSTAT:0});
}

// finish deactivating the Record Mode
function finishDeactivatingRecordMode()
{
	var i;

	if(!recmode_active) return; // if it ain't even active, bail

// do this upfront to have key LED blink handling work properly when indications come back
	recmode_active=false;

// send the required GPIO commands to Recorder if it was defined as non-0
	if(recmode_rcdrlpid!=0)
	{
	// FIXME? Make the message format correspond to the spec?
		host.sendNestCmsg('mGPIO',
		{
			id:'WRNI',
			dLPID:recmode_rcdrlpid,
			bvPINN:[gpio.pin_stop_pulse, gpio.pin_on_lamp, gpio.pin_off_lamp],
			bvVALL:[gpio.pulse, gpio.off, gpio.on]
		});
	}

	if(cur_show_loading) // Record Mode is being switched off due to a show profile change
	{
	// turn key off
		mod.setMasterKeyLed(keys.MstRecord, 0);
		mod.setMasterKeyLed(keys.CmbRecord, 0);
		log.info("Record Mode turned OFF due to Show profile change");
		return;
	}

// Turn Channel assignment to prev. settings
	for(i=1; i<=64; i++)
	{
		ch_data[i].recOnswLock=false;
		if(!ch_data[i].exists) continue;
		if(i in rec_backup.CHxx)
			host.sendChnCmsg(i, rec_backup.CHxx[i]);
		if(ch_data[i].pgm[4]) mod.setChnKeyLed(i, keys.ChnPGM4, 255);
	}

// turn CR headphones and monitors back to previous selections
	host.sendNestCmsg('mCRMO', {id:'WRIN', bSSEL:rec_backup.CRMO.bSSEL});
	host.sendNestCmsg('mCRHP', {id:'WRIN', bSSEL:rec_backup.CRHP.bSSEL, bMODE:rec_backup.CRHP.bMODE, bPVMO:rec_backup.CRHP.bPVMO});
//	host.sendNestCmsg('mSTMO', {id:'WRIN', bSSEL:rec_backup.STMO.bSSEL}); // TODO: Mon2 don't fully support SSEL
	host.sendNestCmsg('mSTMO', {id:'WRIN', bvSTAG:rec_backup.STMO.bvSTAG});

// Switch the rightmost meter back to its previous selection
	host.sendNestCmsg(cfg_surroundMode?'mMET2':'mMET4', {id:'WRIN', bSRSW:rec_backup.MET4.bSRSW});

	rec_backup.CHxx=[];
	rec_backup.CRMO=null;
	rec_backup.CRHP=null;
	rec_backup.STMO=null;
	rec_backup.MET4=null;

// turn key off
	mod.setMasterKeyLed(keys.MstRecord, 0);
	mod.setMasterKeyLed(keys.CmbRecord, 0);
	log.info("Record Mode in now OFF");
}
